# -*- coding: utf-8 -*-
import pygame
from pygame.locals import *

import data

import euclid


class Scene(object):
    def __init__(self):
        self.children = []

    def event(self, ev):
        pass

    def _update(self, dt):
        self.update(dt)
        for c in self.children:
            c._update(dt)

    def update(self, dt):
        pass

    def render(self, target):
        self.draw(target)
        for c in sorted(self.children, key=lambda x: x.z):
            c.render(target)

    def draw(self, target):
        pass

    def to_screen(self, (x, y)):
        return (euclid.Point2(*self.director.size) / 2) + euclid.Point2(x, y)

    def from_screen(self, (x, y)):
        return  euclid.Point2(x, y) - euclid.Point2(*self.director.size) / 2

    def add(self, node):
        self.children.append(node)
        node.parent = self

    def remove(self, node):
        assert node in self.children
        node.parent = None
        self.children.remove(node)


class Node(Scene):
    x = 0
    y = 0
    z = 0

    collides = False

    def to_screen(self, (x, y)):
        x, y = self.parent.to_screen((x + self.x, y + self.y))
        return int(x), int(y)

    def from_screen(self, (x, y)):
        return self.parent.from_screen((x, y)) - (self.x, self.y)


class Dummy(object):
    pass


class Director(object):
    def __init__(self, scene, size=(1024, 768), framerate=60, sounds={}):
        self.size = size
        self.framerate = framerate
        self.scenes = [scene]
        self.sound_res = sounds
        self.sounds = Dummy()

    def run(self):
        pygame.mixer.pre_init(channels=1)
        pygame.init()
        screen = pygame.display.set_mode(self.size)
        print screen.get_bitsize()
        self.framerate, _, _ = pygame.mixer.get_init()

        pygame.mixer.set_num_channels(100)
        pygame.mixer.set_reserved(20)

        for sname, fname in self.sound_res.items():
            print "loading", sname, fname
            setattr(self.sounds, sname, pygame.mixer.Sound(data.load(fname)))

        clock = pygame.time.Clock()
        dt = 0
        while True:
            scene = self.scenes[-1]
            scene.director = self
            events = pygame.event.get()
            for ev in events:
                if ev.type == QUIT:
                    return

            for ev in events:
                scene.event(ev)

            scene._update(dt / 1000.)
            scene.render(screen)

            pygame.display.update()
            dt = clock.tick(self.framerate)

            if not self.scenes:
                return

    def replace(self, scene):
        self.scenes[-1] = scene
        scene.director = self