# -*- coding: utf-8 -*-
import pygame
import numpy

GAIN = 10000


class Node(object):
    def __mul__(self, other):
        return Mul(self, other)

    def __imul__(self, other):
        return Mul(self, other)

    def __add__(self, other):
        return Sum(self, other)

    def __iadd__(self, other):
        return Sum(self, other)


class Const(Node):
    def __init__(self, value):
        self.value = value

    def __call__(self, synth, samplerate, duration):
        samples = samplerate * duration

        array = numpy.zeros(samples) + synth.make(self.value)

        return array


class Osc(Node):
    def __init__(self, freq):
        self.freq = freq

    def __call__(self, synth, samplerate, duration):
        samples = samplerate * duration
        freq = synth.make(self.freq)

        if isinstance(freq, numpy.ndarray):
            phasor = (numpy.cumsum(freq) / samples) * \
                duration * 2 * numpy.pi
        else:
            phasor = numpy.linspace(0,
                duration * freq * 2 * numpy.pi,
                samples)
        return numpy.sin(phasor)


class Square(Node):
    def __init__(self, freq):
        self.freq = freq

    def __call__(self, synth, samplerate, duration):
        samples = samplerate * duration
        freq = synth.make(self.freq)

        if isinstance(freq, numpy.ndarray):
            phasor = (numpy.cumsum(freq) / samples) * \
                duration * 2 * numpy.pi
        else:
            phasor = numpy.linspace(0,
                duration * freq * 2 * numpy.pi,
                samples)
        return (phasor % (2 * numpy.pi) > numpy.pi)


class Triangle(Node):
    def __init__(self, freq):
        self.freq = freq

    def __call__(self, synth, samplerate, duration):
        samples = samplerate * duration
        freq = synth.make(self.freq)

        if isinstance(freq, numpy.ndarray):
            phasor = (numpy.cumsum(freq) / samples) * \
                duration * 2 * numpy.pi
        else:
            phasor = numpy.linspace(0,
                duration * freq * 2 * numpy.pi,
                samples)
        saw = phasor % 1
        return (numpy.abs(saw - 0.5) * 4 - 1)


class Sum(Node):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __call__(self, synth, samplerate, duration):
        left = synth.make(self.left)
        right = synth.make(self.right)

        return left + right


class Clip(object):
    def __init__(self, graph, minimum, maximum):
        self.graph = graph
        self.m = minimum
        self.M = maximum

    def __call__(self, samplerate, duration):
        array = self.graph(samplerate, duration)
        return numpy.clip(array, self.m, self.M)


class Envelope(Node):
    def __init__(self, a, d, s, r):
        self.a = a
        self.d = d
        self.s = s
        self.r = r

    def __call__(self, synth, samplerate, duration):
        attack = numpy.linspace(0,
            1, synth.make(self.a) * samplerate)
        decay = numpy.linspace(1,
            synth.make(self.s),
            synth.make(self.d) * samplerate)
        release = numpy.linspace(
            synth.make(self.s), 0,
            synth.make(self.r) * samplerate)
        sustain = numpy.linspace(synth.make(self.s), synth.make(self.s),
            duration * samplerate - len(attack) - len(release) - len(decay))
        envelope = numpy.concatenate((attack, decay, sustain, release))
        fadelen = int(samplerate * 0.05)
        samples = int(samplerate * duration)

        envelope[samples - fadelen:samples] = numpy.minimum(
            envelope[samples - fadelen:samples],
            numpy.linspace(1, 0, fadelen))

        return envelope[:samples]


class Dump(Node):
    def __init__(self, name):
        self.name = name

    def __call__(self, synth, samplerate, duration):
        array = synth.make(self.name)
        numpy.save(open(self.name + '.npy', "w"), array)
        return array


class Mul(Node):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __call__(self, synth, samplerate, duration):
        left = synth.make(self.left)
        right = synth.make(self.right)

        return left * right


class Stereo(Node):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __call__(self, synth, samplerate, duration):
        left = synth.make(self.left)
        right = synth.make(self.right)

        stereo = numpy.array([left, right])
        return stereo.transpose()


class Synth(object):
    def __init__(self, samplerate):
        self.samplerate = samplerate
        self.dtype = numpy.int16
        self.nodes = {}
        self.duration = None

    def __setitem__(self, name, node):
        self.nodes[name] = node

    def __getitem__(self, key):
        return self.nodes[key]

    def make(self, key):
        if isinstance(key, Node):
            node = key
        elif isinstance(key, str):
            node = self.nodes[key]
        else:
            node = key

        if isinstance(node, Node):
            data = node(self, self.samplerate, self.duration)
            return data
        return node

    def build(self, duration):
        self.duration = duration
        array = self.make('out')
        final = (array * GAIN).astype(self.dtype)
        result = pygame.sndarray.make_sound(final)
        self.duration = None
        return result


def bass(f):
    syn = Synth(f)
    syn["attack"] = 0.1
    syn["decay"] = 0.2
    syn['sustain'] = 0.3
    syn['sustain'] = 0.8
    syn['fmdecay'] = 0.8
    fm = Osc('fmfreq') * 'fm_width'
    mod = Envelope(0.05, 'fmdecay', 0.01, 0.01) * fm
    signal = Osc(mod + 'freq')
    e2 = Envelope('attack', 'decay', 'sustain', 0.2)
    out = e2 * signal

    syn["out"] = out

    return syn


if __name__ == "__main__":
    from mingus.containers.Note import Note
    import random

    pygame.mixer.pre_init(channels=1)
    pygame.init()
    f, _, _ = pygame.mixer.get_init()
    pygame.mixer.set_num_channels(100)
    ch = pygame.mixer.Channel(0)

    notes = 'F C G D A E B'.split(" ")
    syn = bass(f)

    while True:
        events = pygame.event.get()
        if not ch.get_queue():
            #ch.queue(selector.select())
            note = Note(random.choice(notes))
            syn['fm_width'] = 1000
            syn['freq'] = note.to_hertz() / 2
            syn['fmfreq'] = note.to_hertz() / 2 * 4. / 5.

            ch.queue(syn.build(random.choice([0.5, 1, 2])))
